#define _CFFI_

/* We try to define Py_LIMITED_API before including Python.h.

   Mess: we can only define it if Py_DEBUG, Py_TRACE_REFS and
   Py_REF_DEBUG are not defined.  This is a best-effort approximation:
   we can learn about Py_DEBUG from pyconfig.h, but it is unclear if
   the same works for the other two macros.  Py_DEBUG implies them,
   but not the other way around.

   The implementation is messy (issue #350): on Windows, with _MSC_VER,
   we have to define Py_LIMITED_API even before including pyconfig.h.
   In that case, we guess what pyconfig.h will do to the macros above,
   and check our guess after the #include.

   Note that on Windows, with CPython 3.x, you need >= 3.5 and virtualenv
   version >= 16.0.0.  With older versions of either, you don't get a
   copy of PYTHON3.DLL in the virtualenv.  We can't check the version of
   CPython *before* we even include pyconfig.h.  ffi.set_source() puts
   a ``#define _CFFI_NO_LIMITED_API'' at the start of this file if it is
   running on Windows < 3.5, as an attempt at fixing it, but that's
   arguably wrong because it may not be the target version of Python.
   Still better than nothing I guess.  As another workaround, you can
   remove the definition of Py_LIMITED_API here.

   See also 'py_limited_api' in cffi/setuptools_ext.py.
*/
#if !defined(_CFFI_USE_EMBEDDING) && !defined(Py_LIMITED_API)
#  ifdef _MSC_VER
#    if !defined(_DEBUG) && !defined(Py_DEBUG) && !defined(Py_TRACE_REFS) && !defined(Py_REF_DEBUG) && !defined(_CFFI_NO_LIMITED_API)
#      define Py_LIMITED_API
#    endif
#    include <pyconfig.h>
     /* sanity-check: Py_LIMITED_API will cause crashes if any of these
        are also defined.  Normally, the Python file PC/pyconfig.h does not
        cause any of these to be defined, with the exception that _DEBUG
        causes Py_DEBUG.  Double-check that. */
#    ifdef Py_LIMITED_API
#      if defined(Py_DEBUG)
#        error "pyconfig.h unexpectedly defines Py_DEBUG, but Py_LIMITED_API is set"
#      endif
#      if defined(Py_TRACE_REFS)
#        error "pyconfig.h unexpectedly defines Py_TRACE_REFS, but Py_LIMITED_API is set"
#      endif
#      if defined(Py_REF_DEBUG)
#        error "pyconfig.h unexpectedly defines Py_REF_DEBUG, but Py_LIMITED_API is set"
#      endif
#    endif
#  else
#    include <pyconfig.h>
#    if !defined(Py_DEBUG) && !defined(Py_TRACE_REFS) && !defined(Py_REF_DEBUG) && !defined(_CFFI_NO_LIMITED_API)
#      define Py_LIMITED_API
#    endif
#  endif
#endif

#include <Python.h>
#ifdef __cplusplus
extern "C" {
#endif
#include <stddef.h>

/* This part is from file 'cffi/parse_c_type.h'.  It is copied at the
   beginning of C sources generated by CFFI's ffi.set_source(). */

typedef void *_cffi_opcode_t;

#define _CFFI_OP(opcode, arg)   (_cffi_opcode_t)(opcode | (((uintptr_t)(arg)) << 8))
#define _CFFI_GETOP(cffi_opcode)    ((unsigned char)(uintptr_t)cffi_opcode)
#define _CFFI_GETARG(cffi_opcode)   (((intptr_t)cffi_opcode) >> 8)

#define _CFFI_OP_PRIMITIVE       1
#define _CFFI_OP_POINTER         3
#define _CFFI_OP_ARRAY           5
#define _CFFI_OP_OPEN_ARRAY      7
#define _CFFI_OP_STRUCT_UNION    9
#define _CFFI_OP_ENUM           11
#define _CFFI_OP_FUNCTION       13
#define _CFFI_OP_FUNCTION_END   15
#define _CFFI_OP_NOOP           17
#define _CFFI_OP_BITFIELD       19
#define _CFFI_OP_TYPENAME       21
#define _CFFI_OP_CPYTHON_BLTN_V 23   // varargs
#define _CFFI_OP_CPYTHON_BLTN_N 25   // noargs
#define _CFFI_OP_CPYTHON_BLTN_O 27   // O  (i.e. a single arg)
#define _CFFI_OP_CONSTANT       29
#define _CFFI_OP_CONSTANT_INT   31
#define _CFFI_OP_GLOBAL_VAR     33
#define _CFFI_OP_DLOPEN_FUNC    35
#define _CFFI_OP_DLOPEN_CONST   37
#define _CFFI_OP_GLOBAL_VAR_F   39
#define _CFFI_OP_EXTERN_PYTHON  41

#define _CFFI_PRIM_VOID          0
#define _CFFI_PRIM_BOOL          1
#define _CFFI_PRIM_CHAR          2
#define _CFFI_PRIM_SCHAR         3
#define _CFFI_PRIM_UCHAR         4
#define _CFFI_PRIM_SHORT         5
#define _CFFI_PRIM_USHORT        6
#define _CFFI_PRIM_INT           7
#define _CFFI_PRIM_UINT          8
#define _CFFI_PRIM_LONG          9
#define _CFFI_PRIM_ULONG        10
#define _CFFI_PRIM_LONGLONG     11
#define _CFFI_PRIM_ULONGLONG    12
#define _CFFI_PRIM_FLOAT        13
#define _CFFI_PRIM_DOUBLE       14
#define _CFFI_PRIM_LONGDOUBLE   15

#define _CFFI_PRIM_WCHAR        16
#define _CFFI_PRIM_INT8         17
#define _CFFI_PRIM_UINT8        18
#define _CFFI_PRIM_INT16        19
#define _CFFI_PRIM_UINT16       20
#define _CFFI_PRIM_INT32        21
#define _CFFI_PRIM_UINT32       22
#define _CFFI_PRIM_INT64        23
#define _CFFI_PRIM_UINT64       24
#define _CFFI_PRIM_INTPTR       25
#define _CFFI_PRIM_UINTPTR      26
#define _CFFI_PRIM_PTRDIFF      27
#define _CFFI_PRIM_SIZE         28
#define _CFFI_PRIM_SSIZE        29
#define _CFFI_PRIM_INT_LEAST8   30
#define _CFFI_PRIM_UINT_LEAST8  31
#define _CFFI_PRIM_INT_LEAST16  32
#define _CFFI_PRIM_UINT_LEAST16 33
#define _CFFI_PRIM_INT_LEAST32  34
#define _CFFI_PRIM_UINT_LEAST32 35
#define _CFFI_PRIM_INT_LEAST64  36
#define _CFFI_PRIM_UINT_LEAST64 37
#define _CFFI_PRIM_INT_FAST8    38
#define _CFFI_PRIM_UINT_FAST8   39
#define _CFFI_PRIM_INT_FAST16   40
#define _CFFI_PRIM_UINT_FAST16  41
#define _CFFI_PRIM_INT_FAST32   42
#define _CFFI_PRIM_UINT_FAST32  43
#define _CFFI_PRIM_INT_FAST64   44
#define _CFFI_PRIM_UINT_FAST64  45
#define _CFFI_PRIM_INTMAX       46
#define _CFFI_PRIM_UINTMAX      47
#define _CFFI_PRIM_FLOATCOMPLEX 48
#define _CFFI_PRIM_DOUBLECOMPLEX 49
#define _CFFI_PRIM_CHAR16       50
#define _CFFI_PRIM_CHAR32       51

#define _CFFI__NUM_PRIM         52
#define _CFFI__UNKNOWN_PRIM           (-1)
#define _CFFI__UNKNOWN_FLOAT_PRIM     (-2)
#define _CFFI__UNKNOWN_LONG_DOUBLE    (-3)

#define _CFFI__IO_FILE_STRUCT         (-1)


struct _cffi_global_s {
    const char *name;
    void *address;
    _cffi_opcode_t type_op;
    void *size_or_direct_fn;  // OP_GLOBAL_VAR: size, or 0 if unknown
                              // OP_CPYTHON_BLTN_*: addr of direct function
};

struct _cffi_getconst_s {
    unsigned long long value;
    const struct _cffi_type_context_s *ctx;
    int gindex;
};

struct _cffi_struct_union_s {
    const char *name;
    int type_index;          // -> _cffi_types, on a OP_STRUCT_UNION
    int flags;               // _CFFI_F_* flags below
    size_t size;
    int alignment;
    int first_field_index;   // -> _cffi_fields array
    int num_fields;
};
#define _CFFI_F_UNION         0x01   // is a union, not a struct
#define _CFFI_F_CHECK_FIELDS  0x02   // complain if fields are not in the
                                     // "standard layout" or if some are missing
#define _CFFI_F_PACKED        0x04   // for CHECK_FIELDS, assume a packed struct
#define _CFFI_F_EXTERNAL      0x08   // in some other ffi.include()
#define _CFFI_F_OPAQUE        0x10   // opaque

struct _cffi_field_s {
    const char *name;
    size_t field_offset;
    size_t field_size;
    _cffi_opcode_t field_type_op;
};

struct _cffi_enum_s {
    const char *name;
    int type_index;          // -> _cffi_types, on a OP_ENUM
    int type_prim;           // _CFFI_PRIM_xxx
    const char *enumerators; // comma-delimited string
};

struct _cffi_typename_s {
    const char *name;
    int type_index;   /* if opaque, points to a possibly artificial
                         OP_STRUCT which is itself opaque */
};

struct _cffi_type_context_s {
    _cffi_opcode_t *types;
    const struct _cffi_global_s *globals;
    const struct _cffi_field_s *fields;
    const struct _cffi_struct_union_s *struct_unions;
    const struct _cffi_enum_s *enums;
    const struct _cffi_typename_s *typenames;
    int num_globals;
    int num_struct_unions;
    int num_enums;
    int num_typenames;
    const char *const *includes;
    int num_types;
    int flags;      /* future extension */
};

struct _cffi_parse_info_s {
    const struct _cffi_type_context_s *ctx;
    _cffi_opcode_t *output;
    unsigned int output_size;
    size_t error_location;
    const char *error_message;
};

struct _cffi_externpy_s {
    const char *name;
    size_t size_of_result;
    void *reserved1, *reserved2;
};

#ifdef _CFFI_INTERNAL
static int parse_c_type(struct _cffi_parse_info_s *info, const char *input);
static int search_in_globals(const struct _cffi_type_context_s *ctx,
                             const char *search, size_t search_len);
static int search_in_struct_unions(const struct _cffi_type_context_s *ctx,
                                   const char *search, size_t search_len);
#endif

/* this block of #ifs should be kept exactly identical between
   c/_cffi_backend.c, cffi/vengine_cpy.py, cffi/vengine_gen.py
   and cffi/_cffi_include.h */
#if defined(_MSC_VER)
# include <malloc.h>   /* for alloca() */
# if _MSC_VER < 1600   /* MSVC < 2010 */
   typedef __int8 int8_t;
   typedef __int16 int16_t;
   typedef __int32 int32_t;
   typedef __int64 int64_t;
   typedef unsigned __int8 uint8_t;
   typedef unsigned __int16 uint16_t;
   typedef unsigned __int32 uint32_t;
   typedef unsigned __int64 uint64_t;
   typedef __int8 int_least8_t;
   typedef __int16 int_least16_t;
   typedef __int32 int_least32_t;
   typedef __int64 int_least64_t;
   typedef unsigned __int8 uint_least8_t;
   typedef unsigned __int16 uint_least16_t;
   typedef unsigned __int32 uint_least32_t;
   typedef unsigned __int64 uint_least64_t;
   typedef __int8 int_fast8_t;
   typedef __int16 int_fast16_t;
   typedef __int32 int_fast32_t;
   typedef __int64 int_fast64_t;
   typedef unsigned __int8 uint_fast8_t;
   typedef unsigned __int16 uint_fast16_t;
   typedef unsigned __int32 uint_fast32_t;
   typedef unsigned __int64 uint_fast64_t;
   typedef __int64 intmax_t;
   typedef unsigned __int64 uintmax_t;
# else
#  include <stdint.h>
# endif
# if _MSC_VER < 1800   /* MSVC < 2013 */
#  ifndef __cplusplus
    typedef unsigned char _Bool;
#  endif
# endif
#else
# include <stdint.h>
# if (defined (__SVR4) && defined (__sun)) || defined(_AIX) || defined(__hpux)
#  include <alloca.h>
# endif
#endif

#ifdef __GNUC__
# define _CFFI_UNUSED_FN  __attribute__((unused))
#else
# define _CFFI_UNUSED_FN  /* nothing */
#endif

#ifdef __cplusplus
# ifndef _Bool
   typedef bool _Bool;   /* semi-hackish: C++ has no _Bool; bool is builtin */
# endif
#endif

/**********  CPython-specific section  **********/
#ifndef PYPY_VERSION


#if PY_MAJOR_VERSION >= 3
# define PyInt_FromLong PyLong_FromLong
#endif

#define _cffi_from_c_double PyFloat_FromDouble
#define _cffi_from_c_float PyFloat_FromDouble
#define _cffi_from_c_long PyInt_FromLong
#define _cffi_from_c_ulong PyLong_FromUnsignedLong
#define _cffi_from_c_longlong PyLong_FromLongLong
#define _cffi_from_c_ulonglong PyLong_FromUnsignedLongLong
#define _cffi_from_c__Bool PyBool_FromLong

#define _cffi_to_c_double PyFloat_AsDouble
#define _cffi_to_c_float PyFloat_AsDouble

#define _cffi_from_c_int(x, type)                                        \
    (((type)-1) > 0 ? /* unsigned */                                     \
        (sizeof(type) < sizeof(long) ?                                   \
            PyInt_FromLong((long)x) :                                    \
         sizeof(type) == sizeof(long) ?                                  \
            PyLong_FromUnsignedLong((unsigned long)x) :                  \
            PyLong_FromUnsignedLongLong((unsigned long long)x)) :        \
        (sizeof(type) <= sizeof(long) ?                                  \
            PyInt_FromLong((long)x) :                                    \
            PyLong_FromLongLong((long long)x)))

#define _cffi_to_c_int(o, type)                                          \
    ((type)(                                                             \
     sizeof(type) == 1 ? (((type)-1) > 0 ? (type)_cffi_to_c_u8(o)        \
                                         : (type)_cffi_to_c_i8(o)) :     \
     sizeof(type) == 2 ? (((type)-1) > 0 ? (type)_cffi_to_c_u16(o)       \
                                         : (type)_cffi_to_c_i16(o)) :    \
     sizeof(type) == 4 ? (((type)-1) > 0 ? (type)_cffi_to_c_u32(o)       \
                                         : (type)_cffi_to_c_i32(o)) :    \
     sizeof(type) == 8 ? (((type)-1) > 0 ? (type)_cffi_to_c_u64(o)       \
                                         : (type)_cffi_to_c_i64(o)) :    \
     (Py_FatalError("unsupported size for type " #type), (type)0)))

#define _cffi_to_c_i8                                                    \
                 ((int(*)(PyObject *))_cffi_exports[1])
#define _cffi_to_c_u8                                                    \
                 ((int(*)(PyObject *))_cffi_exports[2])
#define _cffi_to_c_i16                                                   \
                 ((int(*)(PyObject *))_cffi_exports[3])
#define _cffi_to_c_u16                                                   \
                 ((int(*)(PyObject *))_cffi_exports[4])
#define _cffi_to_c_i32                                                   \
                 ((int(*)(PyObject *))_cffi_exports[5])
#define _cffi_to_c_u32                                                   \
                 ((unsigned int(*)(PyObject *))_cffi_exports[6])
#define _cffi_to_c_i64                                                   \
                 ((long long(*)(PyObject *))_cffi_exports[7])
#define _cffi_to_c_u64                                                   \
                 ((unsigned long long(*)(PyObject *))_cffi_exports[8])
#define _cffi_to_c_char                                                  \
                 ((int(*)(PyObject *))_cffi_exports[9])
#define _cffi_from_c_pointer                                             \
    ((PyObject *(*)(char *, struct _cffi_ctypedescr *))_cffi_exports[10])
#define _cffi_to_c_pointer                                               \
    ((char *(*)(PyObject *, struct _cffi_ctypedescr *))_cffi_exports[11])
#define _cffi_get_struct_layout                                          \
    not used any more
#define _cffi_restore_errno                                              \
    ((void(*)(void))_cffi_exports[13])
#define _cffi_save_errno                                                 \
    ((void(*)(void))_cffi_exports[14])
#define _cffi_from_c_char                                                \
    ((PyObject *(*)(char))_cffi_exports[15])
#define _cffi_from_c_deref                                               \
    ((PyObject *(*)(char *, struct _cffi_ctypedescr *))_cffi_exports[16])
#define _cffi_to_c                                                       \
    ((int(*)(char *, struct _cffi_ctypedescr *, PyObject *))_cffi_exports[17])
#define _cffi_from_c_struct                                              \
    ((PyObject *(*)(char *, struct _cffi_ctypedescr *))_cffi_exports[18])
#define _cffi_to_c_wchar_t                                               \
    ((_cffi_wchar_t(*)(PyObject *))_cffi_exports[19])
#define _cffi_from_c_wchar_t                                             \
    ((PyObject *(*)(_cffi_wchar_t))_cffi_exports[20])
#define _cffi_to_c_long_double                                           \
    ((long double(*)(PyObject *))_cffi_exports[21])
#define _cffi_to_c__Bool                                                 \
    ((_Bool(*)(PyObject *))_cffi_exports[22])
#define _cffi_prepare_pointer_call_argument                              \
    ((Py_ssize_t(*)(struct _cffi_ctypedescr *,                           \
                    PyObject *, char **))_cffi_exports[23])
#define _cffi_convert_array_from_object                                  \
    ((int(*)(char *, struct _cffi_ctypedescr *, PyObject *))_cffi_exports[24])
#define _CFFI_CPIDX  25
#define _cffi_call_python                                                \
    ((void(*)(struct _cffi_externpy_s *, char *))_cffi_exports[_CFFI_CPIDX])
#define _cffi_to_c_wchar3216_t                                           \
    ((int(*)(PyObject *))_cffi_exports[26])
#define _cffi_from_c_wchar3216_t                                         \
    ((PyObject *(*)(int))_cffi_exports[27])
#define _CFFI_NUM_EXPORTS 28

struct _cffi_ctypedescr;

static void *_cffi_exports[_CFFI_NUM_EXPORTS];

#define _cffi_type(index)   (                           \
    assert((((uintptr_t)_cffi_types[index]) & 1) == 0), \
    (struct _cffi_ctypedescr *)_cffi_types[index])

static PyObject *_cffi_init(const char *module_name, Py_ssize_t version,
                            const struct _cffi_type_context_s *ctx)
{
    PyObject *module, *o_arg, *new_module;
    void *raw[] = {
        (void *)module_name,
        (void *)version,
        (void *)_cffi_exports,
        (void *)ctx,
    };

    module = PyImport_ImportModule("_cffi_backend");
    if (module == NULL)
        goto failure;

    o_arg = PyLong_FromVoidPtr((void *)raw);
    if (o_arg == NULL)
        goto failure;

    new_module = PyObject_CallMethod(
        module, (char *)"_init_cffi_1_0_external_module", (char *)"O", o_arg);

    Py_DECREF(o_arg);
    Py_DECREF(module);
    return new_module;

  failure:
    Py_XDECREF(module);
    return NULL;
}


#ifdef HAVE_WCHAR_H
typedef wchar_t _cffi_wchar_t;
#else
typedef uint16_t _cffi_wchar_t;   /* same random pick as _cffi_backend.c */
#endif

_CFFI_UNUSED_FN static uint16_t _cffi_to_c_char16_t(PyObject *o)
{
    if (sizeof(_cffi_wchar_t) == 2)
        return (uint16_t)_cffi_to_c_wchar_t(o);
    else
        return (uint16_t)_cffi_to_c_wchar3216_t(o);
}

_CFFI_UNUSED_FN static PyObject *_cffi_from_c_char16_t(uint16_t x)
{
    if (sizeof(_cffi_wchar_t) == 2)
        return _cffi_from_c_wchar_t((_cffi_wchar_t)x);
    else
        return _cffi_from_c_wchar3216_t((int)x);
}

_CFFI_UNUSED_FN static int _cffi_to_c_char32_t(PyObject *o)
{
    if (sizeof(_cffi_wchar_t) == 4)
        return (int)_cffi_to_c_wchar_t(o);
    else
        return (int)_cffi_to_c_wchar3216_t(o);
}

_CFFI_UNUSED_FN static PyObject *_cffi_from_c_char32_t(unsigned int x)
{
    if (sizeof(_cffi_wchar_t) == 4)
        return _cffi_from_c_wchar_t((_cffi_wchar_t)x);
    else
        return _cffi_from_c_wchar3216_t((int)x);
}

union _cffi_union_alignment_u {
    unsigned char m_char;
    unsigned short m_short;
    unsigned int m_int;
    unsigned long m_long;
    unsigned long long m_longlong;
    float m_float;
    double m_double;
    long double m_longdouble;
};

struct _cffi_freeme_s {
    struct _cffi_freeme_s *next;
    union _cffi_union_alignment_u alignment;
};

_CFFI_UNUSED_FN static int
_cffi_convert_array_argument(struct _cffi_ctypedescr *ctptr, PyObject *arg,
                             char **output_data, Py_ssize_t datasize,
                             struct _cffi_freeme_s **freeme)
{
    char *p;
    if (datasize < 0)
        return -1;

    p = *output_data;
    if (p == NULL) {
        struct _cffi_freeme_s *fp = (struct _cffi_freeme_s *)PyObject_Malloc(
            offsetof(struct _cffi_freeme_s, alignment) + (size_t)datasize);
        if (fp == NULL)
            return -1;
        fp->next = *freeme;
        *freeme = fp;
        p = *output_data = (char *)&fp->alignment;
    }
    memset((void *)p, 0, (size_t)datasize);
    return _cffi_convert_array_from_object(p, ctptr, arg);
}

_CFFI_UNUSED_FN static void
_cffi_free_array_arguments(struct _cffi_freeme_s *freeme)
{
    do {
        void *p = (void *)freeme;
        freeme = freeme->next;
        PyObject_Free(p);
    } while (freeme != NULL);
}

/**********  end CPython-specific section  **********/
#else
_CFFI_UNUSED_FN
static void (*_cffi_call_python_org)(struct _cffi_externpy_s *, char *);
# define _cffi_call_python  _cffi_call_python_org
#endif


#define _cffi_array_len(array)   (sizeof(array) / sizeof((array)[0]))

#define _cffi_prim_int(size, sign)                                      \
    ((size) == 1 ? ((sign) ? _CFFI_PRIM_INT8  : _CFFI_PRIM_UINT8)  :    \
     (size) == 2 ? ((sign) ? _CFFI_PRIM_INT16 : _CFFI_PRIM_UINT16) :    \
     (size) == 4 ? ((sign) ? _CFFI_PRIM_INT32 : _CFFI_PRIM_UINT32) :    \
     (size) == 8 ? ((sign) ? _CFFI_PRIM_INT64 : _CFFI_PRIM_UINT64) :    \
     _CFFI__UNKNOWN_PRIM)

#define _cffi_prim_float(size)                                          \
    ((size) == sizeof(float) ? _CFFI_PRIM_FLOAT :                       \
     (size) == sizeof(double) ? _CFFI_PRIM_DOUBLE :                     \
     (size) == sizeof(long double) ? _CFFI__UNKNOWN_LONG_DOUBLE :       \
     _CFFI__UNKNOWN_FLOAT_PRIM)

#define _cffi_check_int(got, got_nonpos, expected)      \
    ((got_nonpos) == (expected <= 0) &&                 \
     (got) == (unsigned long long)expected)

#ifdef MS_WIN32
# define _cffi_stdcall  __stdcall
#else
# define _cffi_stdcall  /* nothing */
#endif

#ifdef __cplusplus
}
#endif

/************************************************************/

extern "C" {
    #include "assimp.h"
}

#include <assimp/Importer.hpp>
#include <assimp/Exporter.hpp>
#include <assimp/DefaultLogger.hpp>
#include <assimp/postprocess.h>
#include <assimp/scene.h>

#include <set>
#include <string>
#include <sstream>
#include <unordered_map>

using namespace Assimp;

const auto MAGIC_PNG = "\x89PNG\r\n";
const auto MAGIC_JPEG = "\xFF\xD8\xFF\xE0";

ImageNode* extract_image_nodes(aiScene* pScene);
void import_image_nodes(aiScene* pScene, ImageNode* images);
void add_texture(aiScene* pScene, char* bytes, int bytes_length);

void assimp_import_from_bytes(char* bytes_in, int length_in, int flags_in, Scene* scene_out, int verbose) {
    // Enable verbose mode if asked
    if (verbose) {
        DefaultLogger::create(nullptr, Logger::NORMAL, aiDefaultLogStream_STDOUT);
    }

    unsigned int flags = 0u
        // | aiProcess_FindDegenerates          // Do not use, this can generate points and lines
        // | aiProcess_PreTransformVertices     // A bit too harsh, as it will remove the scene hierarchy
        // | aiProcess_EmbedTextures            // We do that by ourself in python
        | aiProcess_JoinIdenticalVertices
        | aiProcess_Triangulate
        | aiProcess_GenSmoothNormals
        | aiProcess_ImproveCacheLocality
        | aiProcess_RemoveRedundantMaterials
        | aiProcess_FindInvalidData
        | aiProcess_GenUVCoords;

    if (flags_in & FLAG_OPTIMIZE_GRAPH) {
        flags |= aiProcess_OptimizeGraph;
    }
    if (flags_in & FLAG_OPTIMIZE_MESHES) {
        flags |= aiProcess_OptimizeMeshes;
    }
    if (flags_in & FLAG_FIX_INFACING_NORMALS) {
        flags |= aiProcess_FixInfacingNormals;
    }

    Importer importer;
    importer.ReadFileFromMemory(bytes_in, length_in, flags);

    // Free the ownership of the scene from the importer
    auto pScene = importer.GetOrphanedScene();
    scene_out->assimp_scene = pScene;
    if (pScene == nullptr) return;

    // Extract image nodes
    scene_out->images = extract_image_nodes(pScene);
}

int assimp_export_to_bytes(Scene* scene_in, OutputFormat output_format_in, char** bytes_out) {
    auto pScene = reinterpret_cast<aiScene*>(scene_in->assimp_scene);
    Exporter exporter;

    import_image_nodes(pScene, scene_in->images);

    auto outputFormat = output_format_in == OUTPUT_FORMAT_GLB ? "glb2" : "gltf2";
    auto blob = exporter.ExportToBlob(pScene, outputFormat, 0, nullptr);
    if (blob == nullptr) return 0;

    // Copying bytes - as they will be freed
    *bytes_out = new char[blob->size];
    memcpy(*bytes_out, blob->data, blob->size);
    return blob->size;
}

void assimp_free_bytes(char** bytes) {
    delete[] *bytes;
    *bytes = nullptr;
}

void assimp_free_scene(Scene* scene) {
    free(reinterpret_cast<aiScene*>(scene->assimp_scene));

    auto image = scene->images;
    while (image != nullptr) {
        auto nextImage = image->next;
        delete image;
        image = nextImage;
    }
}

//---- Private

ImageNode* extract_image_nodes(aiScene* pScene) {
    ImageNode* images = nullptr;

    std::set<std::string> seenTextures;
    aiString path;
    for (auto matId = 0u; matId < pScene->mNumMaterials; ++matId) {
        auto material = pScene->mMaterials[matId];
 
        for (auto ttId = 1u; ttId < AI_TEXTURE_TYPE_MAX; ++ttId) {
            auto tt = static_cast<aiTextureType>(ttId);
            auto texturesCount = material->GetTextureCount(tt);
 
            for (auto texId = 0u; texId < texturesCount; ++texId) {
                material->GetTexture(tt, texId, &path);
                if (seenTextures.find(path.C_Str()) != seenTextures.end()) continue;

                seenTextures.emplace(path.C_Str()); 
                auto imageNode = new ImageNode();
                imageNode->path = strcpy(new char[path.length + 1], path.C_Str());
                imageNode->next = images;
                images = imageNode;

                // Load bytes of embedded textures
                if (path.data[0] == '*') {
                    // @todo Already embedded textures are not supported yet
                }
            }
        }
    }

    return images;
}

void import_image_nodes(aiScene* pScene, ImageNode* images) {
    std::unordered_map<std::string, std::string> texturesMap;
    std::unordered_map<int, std::string> seenTextures;
    
    // Embed all images
    auto image = images;
    while (image != nullptr) {
        if (image->bytes_length > 0 && image->path[0] != '*') {
            // Images might have different paths at first,
            // they still can point to the same image - as identified by their id.
            if (seenTextures.find(image->id) != std::end(seenTextures)) {
                texturesMap[image->path] = seenTextures[image->id];
            } else {
                add_texture(pScene, image->bytes, image->bytes_length);

                auto embeddedTextureId = pScene->mNumTextures - 1u;
                std::stringstream path;
                path << "*" << embeddedTextureId;
                texturesMap[image->path] = path.str();
                seenTextures[image->id] = path.str();
            }
        }

        image = image->next;
    }

    // Update materials references
    aiString path;
    for (auto matId = 0u; matId < pScene->mNumMaterials; ++matId) {
        auto material = pScene->mMaterials[matId];
 
        for (auto ttId = 1u; ttId < AI_TEXTURE_TYPE_MAX; ++ttId) {
            auto tt = static_cast<aiTextureType>(ttId);
            auto texturesCount = material->GetTextureCount(tt);
 
            for (auto texId = 0u; texId < texturesCount; ++texId) {
                material->GetTexture(tt, texId, &path);
                auto newPath = texturesMap.find(path.C_Str());
                if (newPath == texturesMap.end()) continue;

                path = newPath->second;
                material->AddProperty(&path, AI_MATKEY_TEXTURE(tt, texId));

                // @todo If already embedded texture, update textures table
            }
        }
    }
}

void add_texture(aiScene* pScene, char* bytes, int bytes_length) {
    // Grow the textures table
    auto textureId = pScene->mNumTextures++;
    auto oldTextures = pScene->mTextures;
    pScene->mTextures = new aiTexture*[pScene->mNumTextures];
    memmove(pScene->mTextures, oldTextures, sizeof(aiTexture*) * (pScene->mNumTextures - 1u));

    // Add the new texture
    auto pTexture = new aiTexture();
    pTexture->mHeight = 0; // Means that this is still compressed
    pTexture->mWidth = bytes_length;
    pTexture->pcData = reinterpret_cast<aiTexel*>(bytes);

    if (strncmp(bytes, MAGIC_JPEG, 4) == 0) {
        strcpy(pTexture->achFormatHint, "jpg");
    } else if (strncmp(bytes, MAGIC_PNG, 6) == 0) {
        strcpy(pTexture->achFormatHint, "png");
    }

    pScene->mTextures[textureId] = pTexture;
}


/************************************************************/

static void *_cffi_types[] = {
/*  0 */ _CFFI_OP(_CFFI_OP_FUNCTION, 13), // int()(Scene *, OutputFormat, char * *)
/*  1 */ _CFFI_OP(_CFFI_OP_POINTER, 21), // Scene *
/*  2 */ _CFFI_OP(_CFFI_OP_ENUM, 1), // OutputFormat
/*  3 */ _CFFI_OP(_CFFI_OP_POINTER, 12), // char * *
/*  4 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/*  5 */ _CFFI_OP(_CFFI_OP_FUNCTION, 25), // void()(Scene *)
/*  6 */ _CFFI_OP(_CFFI_OP_NOOP, 1),
/*  7 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/*  8 */ _CFFI_OP(_CFFI_OP_FUNCTION, 25), // void()(char * *)
/*  9 */ _CFFI_OP(_CFFI_OP_NOOP, 3),
/* 10 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 11 */ _CFFI_OP(_CFFI_OP_FUNCTION, 25), // void()(char *, int, int, Scene *, int)
/* 12 */ _CFFI_OP(_CFFI_OP_POINTER, 23), // char *
/* 13 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 7), // int
/* 14 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 7),
/* 15 */ _CFFI_OP(_CFFI_OP_NOOP, 1),
/* 16 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 7),
/* 17 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 18 */ _CFFI_OP(_CFFI_OP_ENUM, 0), // Flag
/* 19 */ _CFFI_OP(_CFFI_OP_POINTER, 20), // ImageNode *
/* 20 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 0), // ImageNode
/* 21 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 1), // Scene
/* 22 */ _CFFI_OP(_CFFI_OP_POINTER, 23), // char const *
/* 23 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 2), // char
/* 24 */ _CFFI_OP(_CFFI_OP_POINTER, 25), // void *
/* 25 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 0), // void
};

static int _cffi_const_FLAG_OPTIMIZE_GRAPH(unsigned long long *o)
{
  int n = (FLAG_OPTIMIZE_GRAPH) <= 0;
  *o = (unsigned long long)((FLAG_OPTIMIZE_GRAPH) | 0);  /* check that FLAG_OPTIMIZE_GRAPH is an integer */
  return n;
}

static int _cffi_const_FLAG_OPTIMIZE_MESHES(unsigned long long *o)
{
  int n = (FLAG_OPTIMIZE_MESHES) <= 0;
  *o = (unsigned long long)((FLAG_OPTIMIZE_MESHES) | 0);  /* check that FLAG_OPTIMIZE_MESHES is an integer */
  return n;
}

static int _cffi_const_FLAG_FIX_INFACING_NORMALS(unsigned long long *o)
{
  int n = (FLAG_FIX_INFACING_NORMALS) <= 0;
  *o = (unsigned long long)((FLAG_FIX_INFACING_NORMALS) | 0);  /* check that FLAG_FIX_INFACING_NORMALS is an integer */
  return n;
}

static int _cffi_const_OUTPUT_FORMAT_GLTF(unsigned long long *o)
{
  int n = (OUTPUT_FORMAT_GLTF) <= 0;
  *o = (unsigned long long)((OUTPUT_FORMAT_GLTF) | 0);  /* check that OUTPUT_FORMAT_GLTF is an integer */
  return n;
}

static int _cffi_const_OUTPUT_FORMAT_GLB(unsigned long long *o)
{
  int n = (OUTPUT_FORMAT_GLB) <= 0;
  *o = (unsigned long long)((OUTPUT_FORMAT_GLB) | 0);  /* check that OUTPUT_FORMAT_GLB is an integer */
  return n;
}

static int _cffi_d_assimp_export_to_bytes(Scene * x0, OutputFormat x1, char * * x2)
{
  return assimp_export_to_bytes(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_assimp_export_to_bytes(PyObject *self, PyObject *args)
{
  Scene * x0;
  OutputFormat x1;
  char * * x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  int result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "assimp_export_to_bytes", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (Scene *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(2), arg1) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (char * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = assimp_export_to_bytes(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, int);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_assimp_export_to_bytes _cffi_d_assimp_export_to_bytes
#endif

static void _cffi_d_assimp_free_bytes(char * * x0)
{
  assimp_free_bytes(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_assimp_free_bytes(PyObject *self, PyObject *arg0)
{
  char * * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { assimp_free_bytes(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_assimp_free_bytes _cffi_d_assimp_free_bytes
#endif

static void _cffi_d_assimp_free_scene(Scene * x0)
{
  assimp_free_scene(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_assimp_free_scene(PyObject *self, PyObject *arg0)
{
  Scene * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (Scene *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { assimp_free_scene(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_assimp_free_scene _cffi_d_assimp_free_scene
#endif

static void _cffi_d_assimp_import_from_bytes(char * x0, int x1, int x2, Scene * x3, int x4)
{
  assimp_import_from_bytes(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_assimp_import_from_bytes(PyObject *self, PyObject *args)
{
  char * x0;
  int x1;
  int x2;
  Scene * x3;
  int x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "assimp_import_from_bytes", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(12), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(12), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, int);
  if (x1 == (int)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, int);
  if (x2 == (int)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (Scene *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, int);
  if (x4 == (int)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { assimp_import_from_bytes(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_assimp_import_from_bytes _cffi_d_assimp_import_from_bytes
#endif

_CFFI_UNUSED_FN
static void _cffi_checkfld__ImageNode(ImageNode *p)
{
  /* only to generate compile-time warnings or errors */
  (void)p;
  { ImageNode * *tmp = &p->next; (void)tmp; }
  { char const * *tmp = &p->path; (void)tmp; }
  (void)((p->id) | 0);  /* check that 'ImageNode.id' is an integer */
  { char * *tmp = &p->bytes; (void)tmp; }
  (void)((p->bytes_length) | 0);  /* check that 'ImageNode.bytes_length' is an integer */
}
struct _cffi_align__ImageNode { char x; ImageNode y; };

_CFFI_UNUSED_FN
static void _cffi_checkfld__Scene(Scene *p)
{
  /* only to generate compile-time warnings or errors */
  (void)p;
  { void * *tmp = &p->assimp_scene; (void)tmp; }
  { ImageNode * *tmp = &p->images; (void)tmp; }
}
struct _cffi_align__Scene { char x; Scene y; };

static const struct _cffi_global_s _cffi_globals[] = {
  { "FLAG_FIX_INFACING_NORMALS", (void *)_cffi_const_FLAG_FIX_INFACING_NORMALS, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "FLAG_OPTIMIZE_GRAPH", (void *)_cffi_const_FLAG_OPTIMIZE_GRAPH, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "FLAG_OPTIMIZE_MESHES", (void *)_cffi_const_FLAG_OPTIMIZE_MESHES, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "OUTPUT_FORMAT_GLB", (void *)_cffi_const_OUTPUT_FORMAT_GLB, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "OUTPUT_FORMAT_GLTF", (void *)_cffi_const_OUTPUT_FORMAT_GLTF, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "assimp_export_to_bytes", (void *)_cffi_f_assimp_export_to_bytes, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 0), (void *)_cffi_d_assimp_export_to_bytes },
  { "assimp_free_bytes", (void *)_cffi_f_assimp_free_bytes, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 8), (void *)_cffi_d_assimp_free_bytes },
  { "assimp_free_scene", (void *)_cffi_f_assimp_free_scene, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 5), (void *)_cffi_d_assimp_free_scene },
  { "assimp_import_from_bytes", (void *)_cffi_f_assimp_import_from_bytes, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 11), (void *)_cffi_d_assimp_import_from_bytes },
};

static const struct _cffi_field_s _cffi_fields[] = {
  { "next", offsetof(ImageNode, next),
            sizeof(((ImageNode *)0)->next),
            _CFFI_OP(_CFFI_OP_NOOP, 19) },
  { "path", offsetof(ImageNode, path),
            sizeof(((ImageNode *)0)->path),
            _CFFI_OP(_CFFI_OP_NOOP, 22) },
  { "id", offsetof(ImageNode, id),
          sizeof(((ImageNode *)0)->id),
          _CFFI_OP(_CFFI_OP_NOOP, 13) },
  { "bytes", offsetof(ImageNode, bytes),
             sizeof(((ImageNode *)0)->bytes),
             _CFFI_OP(_CFFI_OP_NOOP, 12) },
  { "bytes_length", offsetof(ImageNode, bytes_length),
                    sizeof(((ImageNode *)0)->bytes_length),
                    _CFFI_OP(_CFFI_OP_NOOP, 13) },
  { "assimp_scene", offsetof(Scene, assimp_scene),
                    sizeof(((Scene *)0)->assimp_scene),
                    _CFFI_OP(_CFFI_OP_NOOP, 24) },
  { "images", offsetof(Scene, images),
              sizeof(((Scene *)0)->images),
              _CFFI_OP(_CFFI_OP_NOOP, 19) },
};

static const struct _cffi_struct_union_s _cffi_struct_unions[] = {
  { "ImageNode", 20, _CFFI_F_CHECK_FIELDS,
    sizeof(ImageNode), offsetof(struct _cffi_align__ImageNode, y), 0, 5 },
  { "Scene", 21, _CFFI_F_CHECK_FIELDS,
    sizeof(Scene), offsetof(struct _cffi_align__Scene, y), 5, 2 },
};

static const struct _cffi_enum_s _cffi_enums[] = {
  { "Flag", 18, _cffi_prim_int(sizeof(Flag), ((Flag)-1) <= 0),
    "FLAG_OPTIMIZE_GRAPH,FLAG_OPTIMIZE_MESHES,FLAG_FIX_INFACING_NORMALS" },
  { "OutputFormat", 2, _cffi_prim_int(sizeof(OutputFormat), ((OutputFormat)-1) <= 0),
    "OUTPUT_FORMAT_GLTF,OUTPUT_FORMAT_GLB" },
};

static const struct _cffi_typename_s _cffi_typenames[] = {
  { "Flag", 18 },
  { "ImageNode", 20 },
  { "OutputFormat", 2 },
  { "Scene", 21 },
};

static const struct _cffi_type_context_s _cffi_type_context = {
  _cffi_types,
  _cffi_globals,
  _cffi_fields,
  _cffi_struct_unions,
  _cffi_enums,
  _cffi_typenames,
  9,  /* num_globals */
  2,  /* num_struct_unions */
  2,  /* num_enums */
  4,  /* num_typenames */
  NULL,  /* no includes */
  26,  /* num_types */
  0,  /* flags */
};

#ifdef __GNUC__
#  pragma GCC visibility push(default)  /* for -fvisibility= */
#endif

#ifdef PYPY_VERSION
PyMODINIT_FUNC
_cffi_pypyinit__assimp(const void *p[])
{
    p[0] = (const void *)0x2601;
    p[1] = &_cffi_type_context;
#if PY_MAJOR_VERSION >= 3
    return NULL;
#endif
}
#  ifdef _MSC_VER
     PyMODINIT_FUNC
#  if PY_MAJOR_VERSION >= 3
     PyInit__assimp(void) { return NULL; }
#  else
     init_assimp(void) { }
#  endif
#  endif
#elif PY_MAJOR_VERSION >= 3
PyMODINIT_FUNC
PyInit__assimp(void)
{
  return _cffi_init("yoga.model._assimp", 0x2601, &_cffi_type_context);
}
#else
PyMODINIT_FUNC
init_assimp(void)
{
  _cffi_init("yoga.model._assimp", 0x2601, &_cffi_type_context);
}
#endif

#ifdef __GNUC__
#  pragma GCC visibility pop
#endif
